/* drive.c
 * Designed to run on Create Command Module
 *
 * The basic architecture of this program can be re-used to easily 
 * write a wide variety of Create control programs.  All sensor values
 * are polled in the background (using the serial rx interrupt) and 
 * stored in the sensors array as long as the function 
 * delayAndUpdateSensors() is called periodically.  Users can send commands
 * directly a byte at a time using byteTx() or they can use the 
 * provided functions, such as baud() and drive().
 */



// Includes
#include <avr/interrupt.h>
#include <avr/io.h>
#include <stdlib.h>
#include "oi.h"


// Constants
#define RESET_SONG 0
#define START_SONG 1
#define BUMP_SONG  2
#define END_SONG   3
#define JOHN_SONG  4


// Global variables
volatile uint8_t sensors_flag = 0;
volatile uint8_t sensors_index = 0;
volatile uint8_t sensors_in[Sen6Size];
volatile uint8_t sensors[Sen6Size];
volatile int16_t distance = 0;
volatile int16_t angle = 0;




// Functions
void byteTx(uint8_t value);
void delay_cycles(unsigned long int cycles);
void delayAndUpdateSensors(unsigned int time_ms);
void initialize(void);
void powerOnRobot(void);
void baud(uint8_t baud_code);
void drive(int16_t velocity, int16_t radius, int16_t distance);
void go_distance(int16_t distance);
void rotate_angle(void);
void defineSongs(void);

int main (void) 
{
  uint8_t leds_cnt = 99;
  uint8_t leds_state = 0;
  uint8_t leds_on = 1;

  int16_t turn_angle = 0;
  uint8_t turn_dir = 1;
  uint8_t turning = 0;
  uint8_t backing_up = 0;


  // Set up Create and module
  initialize();
  LEDBothOff;
  powerOnRobot();
  byteTx(CmdStart);
  baud(Baud28800);
  defineSongs();
  byteTx(CmdControl);
  byteTx(CmdFull);

  // Stop just as a precaution
  drive(0, RadStraight, 0);

  // Play the reset song and wait while it plays
  byteTx(CmdPlay);
  byteTx(JOHN_SONG);
  delayAndUpdateSensors(750);


  for(;;)//put your stuff in here
  {
/*drive(300, RadStraight);
delay_cycles(65000);
drive(0, RadStraight);
delay_cycles(65000);*/

drive(200, RadCCW, 90);
delay_cycles(65530);
//delay_cycles(65530);
//drive(200, RadCW, 144);
//delay_cycles(65530);

/*go_distance();
delay_cycles(65000);
rotate_angle();
delay_cycles(65000);*/

  delayAndUpdateSensors(200);
  }
}




// Serial receive interrupt to store sensor values
SIGNAL(SIG_USART_RECV)
{
  uint8_t temp;


  temp = UDR0;

  if(sensors_flag)
  {
    sensors_in[sensors_index++] = temp;
    if(sensors_index >= Sen6Size)
      sensors_flag = 0;
  }
}


// Transmit a byte over the serial port
void byteTx(uint8_t value)
{
  while(!(UCSR0A & _BV(UDRE0))) ;
  UDR0 = value;
}



//************DELAY FUNCTIONS************
//wait for X amount of cycles (23 cycles is about .992 milliseconds)
//		this was determined by using an oscilloscope
//to calculate: 40/.992*(time in milliseconds) = number of cycles
//or (number of cycles)*.992/40 = time in milliseconds
//ATmega168 - 28400 cycles per second (28.4 cycles per millisecond)
void delay_cycles(unsigned long int cycles)
	{
	while(cycles > 0)
		cycles--;
	}
//***************************************


// Delay for the specified time in ms and update sensor values
void delayAndUpdateSensors(uint16_t time_ms)
{
  uint8_t temp;

    if(!sensors_flag)
    {
      for(temp = 0; temp < Sen6Size; temp++)
        sensors[temp] = sensors_in[temp];

      // Update running totals of distance and angle
      distance += (int)((sensors[SenDist1] << 8) | sensors[SenDist0]);
      angle += (int)((sensors[SenAng1] << 8) | sensors[SenAng0]);

      byteTx(CmdSensors);
      byteTx(6);
      sensors_index = 0;
      sensors_flag = 1;
  	}
	delay_cycles(time_ms);
}




// Initialize the Mind Control's ATmega168 microcontroller
void initialize(void)
{
  cli();

/*The Command Modules
ePorts expose pins B 0-3 and C 0-5. See the Hardware
Reference in chapter 8 for a table of which microcontroller
pins are connected to which ePort pins.*/

  // Set I/O pins
  DDRB = 0x10;
  PORTB = 0xCF;
  DDRC = 0x00;
  PORTC = 0xFF;
  DDRD = 0xE6;
  PORTD = 0x7D;

/*
  // Set up timer 1 to generate an interrupt every 1 ms
  TCCR1A = 0x00;
  TCCR1B = (_BV(WGM12) | _BV(CS12));
  OCR1A = 71;
  TIMSK1 = _BV(OCIE1A);
  */

  // Set up the serial port with rx interrupt
  UBRR0 = 19;
  UCSR0B = (_BV(RXCIE0) | _BV(TXEN0) | _BV(RXEN0));
  UCSR0C = (_BV(UCSZ00) | _BV(UCSZ01));

  // Turn on interrupts
  //sei();
}




void powerOnRobot(void)
{
  // If Create's power is off, turn it on
  if(!RobotIsOn)
  {
      while(!RobotIsOn)
      {
          RobotPwrToggleLow;
          delay_cycles(500);  // Delay in this state
          RobotPwrToggleHigh;  // Low to high transition to toggle power
          delay_cycles(100);  // Delay in this state
          RobotPwrToggleLow;
      }
      delay_cycles(350);  // Delay for startup
  }
}




// Switch the baud rate on both Create and module
void baud(uint8_t baud_code)
{
  if(baud_code <= 11)
  {
    byteTx(CmdBaud);
    UCSR0A |= _BV(TXC0);
    byteTx(baud_code);
    // Wait until transmit is complete
    while(!(UCSR0A & _BV(TXC0))) ;

    cli();

    // Switch the baud rate register
    if(baud_code == Baud115200)
      UBRR0 = Ubrr115200;
    else if(baud_code == Baud57600)
      UBRR0 = Ubrr57600;
    else if(baud_code == Baud38400)
      UBRR0 = Ubrr38400;
    else if(baud_code == Baud28800)
      UBRR0 = Ubrr28800;
    else if(baud_code == Baud19200)
      UBRR0 = Ubrr19200;
    else if(baud_code == Baud14400)
      UBRR0 = Ubrr14400;
    else if(baud_code == Baud9600)
      UBRR0 = Ubrr9600;
    else if(baud_code == Baud4800)
      UBRR0 = Ubrr4800;
    else if(baud_code == Baud2400)
      UBRR0 = Ubrr2400;
    else if(baud_code == Baud1200)
      UBRR0 = Ubrr1200;
    else if(baud_code == Baud600)
      UBRR0 = Ubrr600;
    else if(baud_code == Baud300)
      UBRR0 = Ubrr300;

    //sei();

    delay_cycles(1000);
  }
}




// Send Create drive commands in terms of velocity and radius
void drive(int16_t velocity, int16_t radius, int16_t distance)
{
  byteTx(CmdDrive);
  byteTx((uint8_t)((velocity >> 8) & 0x00FF));
  byteTx((uint8_t)(velocity & 0x00FF));
  byteTx((uint8_t)((radius >> 8) & 0x00FF));
  byteTx((uint8_t)(radius & 0x00FF));
  byteTx(WaitForAngle);
  byteTx((uint8_t)((distance >> 8) & 0x00FF));
  byteTx((uint8_t)(distance & 0x00FF));
}

void go_distance(int16_t distance)//go straight for 400mm
	{
	//go straight for 400mm
	byteTx(CmdDrive);
	byteTx(1);
	byteTx(44);
	byteTx(128);
	byteTx(0);
	byteTx(156);
	byteTx(1);
	byteTx(144);
	byteTx(153);
	/*
  byteTx(CmdDrive);
  byteTx((uint8_t)((velocity >> 8) & 0x00FF));
  byteTx((uint8_t)(velocity & 0x00FF));
  byteTx((uint8_t)((radius >> 8) & 0x00FF));
  byteTx((uint8_t)(radius & 0x00FF));
  byteTx(WaitForDistance);
  byteTx((uint8_t)((distance >> 8) & 0x00FF));
  byteTx((uint8_t)(distance & 0x00FF));*/
	}

void rotate_angle(void)//rotate CCW 90 degrees
	{
	//rotate CCW 90 degrees
	byteTx(152);//start script
	byteTx(9);//# of bytes after this line
	byteTx(CmdDrive);
	byteTx(1);
	byteTx(44);
	byteTx(0);
	byteTx(1);
	byteTx(157);
	byteTx(0);
	byteTx(90);
	byteTx(153);
	}

// Define songs to be played later
void defineSongs(void)
{
  // Reset song
  byteTx(CmdSong);
  byteTx(RESET_SONG);
  byteTx(4);
  byteTx(60);
  byteTx(6);
  byteTx(72);
  byteTx(6);
  byteTx(84);
  byteTx(6);
  byteTx(96);
  byteTx(6);

  // Start song
  byteTx(CmdSong);
  byteTx(START_SONG);
  byteTx(6);
  byteTx(69);
  byteTx(18);
  byteTx(72);
  byteTx(12);
  byteTx(74);
  byteTx(12);
  byteTx(72);
  byteTx(12);
  byteTx(69);
  byteTx(12);
  byteTx(77);
  byteTx(24);

  // Bump song
  byteTx(CmdSong);
  byteTx(BUMP_SONG);
  byteTx(2);
  byteTx(74);
  byteTx(12);
  byteTx(59);
  byteTx(24);

  // End song
  byteTx(CmdSong);
  byteTx(END_SONG);
  byteTx(5);
  byteTx(77);
  byteTx(18);
  byteTx(74);
  byteTx(12);
  byteTx(72);
  byteTx(12);
  byteTx(69);
  byteTx(12);
  byteTx(65);
  byteTx(24);

  // John Song
  byteTx(CmdSong);
  byteTx(JOHN_SONG);
  byteTx(5);
  byteTx(31);
  byteTx(10);
  byteTx(43);
  byteTx(10);
  byteTx(55);
  byteTx(10);
  byteTx(43);
  byteTx(10);
  byteTx(31);
  byteTx(20);

  // John Song
  byteTx(CmdSong);
  byteTx(JOHN_SONG);
  byteTx(15);
  byteTx(31);
  byteTx(10);
  byteTx(43);
  byteTx(10);
  byteTx(55);
  byteTx(10);
  byteTx(43);
  byteTx(10);
  byteTx(31);
  byteTx(20);
  byteTx(31);
  byteTx(10);
  byteTx(43);
  byteTx(10);
  byteTx(55);
  byteTx(10);
  byteTx(43);
  byteTx(10);
  byteTx(31);
  byteTx(20);
  byteTx(43);
  byteTx(10);
  byteTx(55);
  byteTx(10);
  byteTx(67);
  byteTx(10);
  byteTx(55);
  byteTx(10);
  byteTx(43);
  byteTx(20);

}

